#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "I2cMaster.h"
#include "global.h"
#include "yacrc6686.h"
#include "init6686.h"
#include "ui.h"
#include "nv_memory.h"

extern volatile int8_t nLRot;
extern volatile int8_t nRRot;

uint8_t nFMCEQ;         // FM channel equalizer, 0=off, 1=on
uint8_t nFMEMS;         // FM enhanced multipath suppression, 0=off, 1=on
uint8_t nDeemphasis;    // FM de-emphasis, 0=off, 1=50us, 2=75us

int8_t nRSSI, nRSSI_Last, nRSSI_Disp;  // Received signal strength, in dBuv
int8_t nSNR, nSNR_Last, nSNR_Disp;     // Signal noise ratio, in dB
bool bSTIN;  // FM stereo indicator

// RF Modes
uint8_t nRFMode;  // FM/AM

// Volume
uint8_t nVolume; // Audio volume control, 0-29

// Bands
uint8_t nBand;  // 0-3:LW/MW/SW/FM
const uint8_t nBandMode[NUM_BANDS] = { RFMODE_AM, RFMODE_AM, RFMODE_AM, RFMODE_FM };  // Band RF mode
extern const int32_t nBandFMin[NUM_BANDS] = { 144, 522, 2300, 65000 };
extern const int32_t nBandFMax[NUM_BANDS] = { 288, 1710, 27000, 108000 };
extern const int16_t nBandStep[NUM_BANDS][NUM_STEPS] = { { 9, 1, 45 },{ 9, 1, 90 },{ 5, 1, 500 },{ 100, 10, 500 } };
int32_t nBandFreq[NUM_BANDS] = { 144,522,2300,87600 };  // Band default frequency

// Steps
uint8_t nStepIdx;    // Step index for current band

// Filters
uint8_t nFMFilter;   // Current FIR filter index for FM
uint8_t nAMFilter;   // Current FIR filter index for AM

uint8_t nBacklightAdj;         // LCD backlight value, 0-255
uint8_t nBacklightKeep;        // LCD backlight auto keep seconds, 0-255, 0 for always on
bool bLCDOff = false;          // true for LCD is off
uint32_t nBacklightTimer;      // LCD backlight auto keep timer in ms

uint32_t timer = 0;            // RSSI, SNR & FM stereo indicator display timer

uint16_t nAutoSyncBits;
uint16_t nAutoSyncChs;
uint16_t nAutoSyncFreqs;
uint32_t nAutoSyncTimer;       // Auto sync to NV memory timer

bool bBandChangeOnly = false;
bool bDelayedCheck;
uint32_t nDelayedTimer;
uint8_t nDelayedSmooth;

uint16_t M_FMFilter[] = { 0,56,64,72,84,97,114,133,151,168,184,200,217,236,254,287,311 };
uint8_t M_AMFilter[] = { 3,4,6,8 };
int8_t M_Volume[] = { -60,-56,-52,-48,-46,-44,-42,-40,-38,-36,-34,-32,-30,-28,-26,-24,-22,-20,-18,-16,-14,-12,-10,-8,-6,-4,-2,0,2,3 };

TwiMaster i2c(false);

void Set_Cmd(uint8_t mdl, uint8_t cmd, int len, ...)
{
	uint16_t temp;
	va_list vArgs;
	va_start(vArgs, len);
	i2c.start(DSP_I2C | I2C_WRITE);
	i2c.write(mdl);
	i2c.write(cmd);
	i2c.write(0x01);
	for (uint8_t i = 0; i< len; i++)
	{
		temp = va_arg(vArgs, uint16_t);
		i2c.write((uint8_t)(temp >> 8));
		i2c.write((uint8_t)temp);
	}
	i2c.stop();
	va_end(vArgs);
}

void Get_Cmd(uint8_t mdl, uint8_t cmd, uint16_t *receive, uint8_t len)
{
	i2c.start(DSP_I2C | I2C_WRITE);
	i2c.write(mdl);
	i2c.write(cmd);
	i2c.write(0x01);
	i2c.restart(DSP_I2C | I2C_READ);
	for (uint8_t i = 0; i < len; i++)
	{
		receive[i] = (((uint16_t)i2c.read(false)) << 8);
		receive[i] |= i2c.read(i == (len - 1));
	}
	i2c.stop();
}

void dsp_write_data(const uint8_t* data)
{
	uint8_t *pa = (uint8_t *)data;
	uint8_t len, i;
	for (;;)
	{
		len = pgm_read_byte_near(pa++);
		if (!len)
			break;
		if (len & 0x80)  // Delay
			delay(pgm_read_byte_near(pa++));
		else
		{
			i2c.start(DSP_I2C | I2C_WRITE);
			for (i = 0; i < len; i++)
				i2c.write(pgm_read_byte_near(pa++));
			i2c.stop();
		}
	}
}

void SetVolume(uint8_t v30)
{
	uint8_t i;
	static uint8_t vLast = 0;
	if (v30 == vLast)
		return;
	Set_Cmd(48, 10, 1,  10 * (int16_t)M_Volume[v30]);
	vLast = v30;
}

void NextFilter(void)
{
	if (nRFMode == RFMODE_FM)
		nFMFilter = (nFMFilter + 1) % (NUM_FM_FILTERS);
	else  // RFMODE_AM
		nAMFilter = (nAMFilter + 1) % (NUM_AM_FILTERS);
}

void SetFilter(bool bNeedSync)
{
	if (nRFMode == RFMODE_FM)
	{
		if (nFMFilter == 0)
			Set_Cmd(32, 10, 4, 1, 2360, 1000, 1000);
		else
			Set_Cmd(32, 10, 4, 0, 10 * M_FMFilter[nFMFilter], 1000, 1000);
		if (bNeedSync)
			AddSyncBits(NEEDSYNC_FMFILTER);
	}
	else
	{
		Set_Cmd(33, 10, 2, 0, 10 * M_AMFilter[nAMFilter]);
		if (bNeedSync)
			AddSyncBits(NEEDSYNC_AMFILTER);
	}
}

void SetRFCtrlReg(void)
{
	SetFilter(false);
	if (nRFMode == RFMODE_FM)
	{
		Set_Cmd(32, 20, 1, nFMEMS);
		Set_Cmd(32, 22, 1, nFMCEQ);
		if(nDeemphasis == 0)
			Set_Cmd(32, 31, 1, 0);
		else if (nDeemphasis == 1)
			Set_Cmd(32, 31, 1, 500);
		else if (nDeemphasis == 2)
			Set_Cmd(32, 31, 1, 750);
	}
}

void CheckVolume(void)
{
	int8_t i8;
	if ((i8 = GetLRot()))
	{
		nVolume = constrain(nVolume + i8, MIN_VOL, MAX_VOL);
		SetVolume(nVolume);
		AddSyncBits(NEEDSYNC_VOL);
		CheckUpdateAlt(ALT_VOL);  // Show volume for a period
	}
}

void TuneFreq(int32_t freq)
{
	Set_Cmd(32 + nRFMode, 1, 2, 1, (nRFMode == 1) ? freq : freq / 10);
	bDelayedCheck = true;
	nDelayedTimer = millis() + 30;
	nDelayedSmooth = 4;
	if (!bBandChangeOnly)
	{
		nAutoSyncFreqs |= (1 << nBand);
		nAutoSyncTimer = millis();
	}
}

void AdjFreq(bool bCurrStep)
{
	int32_t nFreq;
	int16_t nStep;
	nFreq = nBandFreq[nBand];
	if (bCurrStep)
		nStep = nBandStep[nBand][nStepIdx];
	else
		nStep = nBandStep[nBand][0];
	if (nFreq < nBandFMin[nBand])
		nFreq = nBandFMax[nBand];
	else if (nFreq > nBandFMax[nBand])
		nFreq = nBandFMin[nBand];
	// Adjust frequency to integer multiples of step value
	nFreq = ((nFreq + (nStep >> 1)) / nStep) * nStep;
	if (nFreq < nBandFMin[nBand])  // Check min margin
		nFreq += nStep;
	else if (nFreq > nBandFMax[nBand])  // Check max margin
		nFreq -= nStep;
	nBandFreq[nBand] = nFreq;
}

void TuneFreqDisp(void)
{
	int32_t f;
	f = nBandFreq[nBand];
	TuneFreq(f);
	if (nBand >= BAND_FM)
	{
		LCDXYIntLen(FREQ_X, FREQ_Y, f / 1000, 3);
		LCDXYChar(FREQ_X + 3, FREQ_Y, '.');
		LCDXYUIntLenZP(FREQ_X + 4, FREQ_Y, f % 1000, 3);
	}
	else
	{
		LCDXYIntLen(FREQ_X, FREQ_Y, f, 6);
		LCDXYChar(FREQ_X + 6, FREQ_Y, 'K');
	}
}

void ProcBand(uint8_t nBd)
{
	if (nStepIdx)
	{
		nStepIdx = 0;  // Default band step
		AddSyncBits(NEEDSYNC_STEPIDX);
	}
	nBand = nBd;
	SetRFMode();
	bBandChangeOnly = true;  // Change band only, don't sync band frequency to NV memory
	TuneFreqDisp();
	bBandChangeOnly = false;
	SetRFCtrlReg();  // Set RF mode related regs
	AddSyncBits(NEEDSYNC_BAND);
}

void ProcStepFilter(uint8_t nKey)
{
	switch (nKey) {
	case KEY_STEP:
		nStepIdx = (nStepIdx + 1) % NUM_STEPS;
		AddSyncBits(NEEDSYNC_STEPIDX);
		break;

	case KEY_STEP | KEY_LONGPRESS:   // Default step
		if (nStepIdx != 0)
			nStepIdx = 0;
		else
			nStepIdx = NUM_STEPS - 1;
		AddSyncBits(NEEDSYNC_STEPIDX);
		break;

	case KEY_FILTER:
		NextFilter();
		SetFilter(true);
		break;

	case KEY_FILTER | KEY_LONGPRESS:// Default filter
		if (nRFMode == RFMODE_FM)
			nFMFilter = DEF_FM_FILTER;
		else
			nAMFilter = DEF_AM_FILTER;
		SetFilter(true);
		break;
	}
}  // void ProcStepFilter(uint8_t nKey)

void GetRFStatReg(void)
{
	uint16_t sts[3];
	Get_Cmd(32 + nRFMode, 128, sts, 3);
	nRSSI = sts[1] / 10;
	uint8_t REG_USN = sts[2] / 10;
	Get_Cmd(32 + nRFMode, 133, sts, 1);
	bSTIN = (sts[0] >> 15) & 1;

	if (nRFMode == RFMODE_FM)  // In FM mode, SNR is NOT calibrated accurately. For reference only
		nSNR = int(0.46222375 * (float)nRSSI - 0.082495118 * (float)REG_USN) + 10;  // Emprical formula
	else  // AM
		nSNR = -((int8_t)REG_USN);
}

void SetRFMode(void)
{
	if (nBandMode[nBand] == RFMODE_FM)
		nRFMode = RFMODE_FM;
	else
		nRFMode = RFMODE_AM;
	AddSyncBits(NEEDSYNC_RFMODE);
}

void AddSyncBits(uint16_t SyncBit)
{
	nAutoSyncBits |= SyncBit;
	nAutoSyncTimer = millis();
}


void setup(void)
{
	uint16_t status;
	Get_Cmd(64, 128, &status, 1);
	if (status < 2)
		dsp_write_data(DSP_INIT);
	pinMode(SH1A_PIN, INPUT_PULLUP);
	pinMode(SH1B_PIN, INPUT_PULLUP);
	pinMode(SH2A_PIN, INPUT_PULLUP);
	pinMode(SH2B_PIN, INPUT_PULLUP);
#ifdef DEBUG
	Serial.begin(115200);
#endif
	LCDInit();
	LCDXYStr_P(4, 0, (PGM_P)F("YACRC6686"));
	delay(500);
	attachInterrupt(0, intLRotEnc, CHANGE);
	attachInterrupt(1, intRRotEnc, CHANGE);
	if (analogRead(0) < ADC_NONE / 2)
		TestRotKey();
	NVMGetArgs();
	LCDClr();
	nLRot = 0;
	nRRot = 0;
	TuneFreqDisp(); // Tune to default frequency and display
	nBacklightTimer = millis();
	SetVolume(nVolume); // Set audio volume
	SetRFCtrlReg();
	analogWrite(LCD_BKLT_PIN, 255);  // Set LCD backlight PWM
	LCDUpdate();

	nAutoSyncBits = 0;
	nAutoSyncChs = 0;
	nAutoSyncFreqs = 0;
	nAutoSyncTimer = 1000L * 3600 * 24;  // Auto sync to NV memory timer
}  // void setup(void)

void loop(void)
{
	int8_t i8;
	uint8_t nKey;
	uint16_t u16;
	CheckVolume();
	if (bDelayedCheck && millis() >= nDelayedTimer)
	{
		bDelayedCheck = false;
		CheckUpdateSig();  // Update RSSI, SNR & FM stereo indicator
		timer = millis() - (TIMER_INTERVAL >> 2);
	}
	else if ((millis() - timer) >= TIMER_INTERVAL)
	{
		CheckUpdateSig();  // Update RSSI, SNR & FM stereo indicator
		timer = millis();
		CheckUpdateAlt(ALT_AUTO);
		if (nBacklightKeep && !bLCDOff)
		{
			if ((timer - nBacklightTimer) >= (uint32_t)nBacklightKeep * 1000)
			{
				analogWrite(LCD_BKLT_PIN, 0);  // Set LCD backlight off
				bLCDOff = true;
			}
		}
		if ((timer - nAutoSyncTimer) >= TIMER_AUTOSYNC)
		{
			if (nAutoSyncBits)
			{
				if (nAutoSyncBits & NEEDSYNC_BAND)
					NV_write_byte((uint8_t *)NVMADDR_BAND, nBand);

				if (nAutoSyncBits & NEEDSYNC_VOL)
					NV_write_byte((uint8_t *)NVMADDR_VOL, nVolume);

				if (nAutoSyncBits & NEEDSYNC_RFMODE)
					NV_write_byte((uint8_t *)NVMADDR_RFMODE, nRFMode);

				if (nAutoSyncBits & NEEDSYNC_STEPIDX)
					NV_write_byte((uint8_t *)NVMADDR_STEPIDX, nStepIdx);

				if (nAutoSyncBits & NEEDSYNC_FMFILTER)
					NV_write_byte((uint8_t *)NVMADDR_FMFILTER, nFMFilter);

				if (nAutoSyncBits & NEEDSYNC_AMFILTER)
					NV_write_byte((uint8_t *)NVMADDR_AMFILTER, nAMFilter);

				if (nAutoSyncBits & NEEDSYNC_MISC1)
					NV_write_byte((uint8_t *)NVMADDR_MISC1, (nDeemphasis << 6));

				if (nAutoSyncBits & NEEDSYNC_MISC2)
					NV_write_byte((uint8_t *)NVMADDR_MISC2, (nFMEMS << 7) | (nFMCEQ << 5));

				if (nAutoSyncBits & NEEDSYNC_MISC3)
					NV_write_byte((uint8_t *)NVMADDR_MISC3, 0);

				nAutoSyncBits = 0;
			}

			if (nAutoSyncFreqs)
			{
				for (i8 = 0; i8 < NUM_BANDS; i8++)
				{
					if (nAutoSyncFreqs & (1 << i8))
					{
						if (i8 <= BAND_SW)
							u16 = (uint16_t)(nBandFreq[i8] - nBandFMin[i8]);
						else
							u16 = (uint16_t)((nBandFreq[i8] - nBandFMin[i8]) / 5);
						NV_write_word((uint16_t *)NVMADDR_BANDFREQ + i8, u16);
					}
				}
				nAutoSyncFreqs = 0;
			}
			nAutoSyncTimer = timer + 1000L * 3600 * 24;
		}
	}
	if ((i8 = GetRRot()))
	{
		nBandFreq[nBand] += i8 * nBandStep[nBand][nStepIdx];
		AdjFreq();
		TuneFreqDisp();
	}
	if ((nKey = GetKey()))
	{
		switch (nKey)
		{
			case KEY_LROT:
			    Menu(MID_OPTION);
			    break;

			//case KEY_LROT | KEY_LONGPRESS:  // Toggle mute/unmute
			//    if(bMuted)
			//        SetVolume(nVolume);  // Unmute
			//    else
			//        SetVolume(0);  // Mute
			//    bMuted = !bMuted;
			//    break;

			case KEY_RROT:
			    Menu(MID_FREQUENCY);
			    break;

			//case KEY_RROT | KEY_LONGPRESS:
			//    AddDelCh();
			//    LCDUpdate();
			//    break;

			//case KEY_TUNE:
			//    nTuneType = (nTuneType + 1) % NUM_TYPES;
			//    if(nTuneType == TYPE_CH)
			//        SeekCh(0);
			//    AddSyncBits(NEEDSYNC_TUNE);
			//    break;

			//     case KEY_TUNE | KEY_LONGPRESS:
			//         if(nTuneType != TYPE_FREQ)
			//             nTuneType = TYPE_FREQ;
			//         else
			//{
			//             nTuneType = TYPE_CH;
			//             SeekCh(0);
			//         }
			//         AddSyncBits(NEEDSYNC_TUNE);
			//         break;

		case KEY_BAND:
			ProcBand((nBand + 1) % NUM_BANDS);
			break;

		default:  // Include KEY_STEP, KEY_STEP | KEY_LONGPRESS, KEY_FILTER, KEY_FILTER | KEY_LONGPRESS
			ProcStepFilter(nKey);
			break;
		}
		LCDUpdate();
	}
}  // void loop(void)
